@@ -19,7 +19,8 @@ from wsproto.utilities import generate_accept_token, split_comma_header
 
 from .events import Body, Data, EndBody, EndData, Event, Request, Response, StreamClosed
 from ..config import Config
-from ..utils import build_and_validate_headers, suppress_body, UnexpectedMessage
+from ..typing import ASGIFramework, Context
+from ..utils import build_and_validate_headers, suppress_body, UnexpectedMessage, valid_server_name
 
 
 class ASGIWebsocketState(Enum):
@@ -135,26 +136,28 @@ class WebsocketBuffer:
 class WSStream:
     def __init__(
         self,
+        app: ASGIFramework,
         config: Config,
+        context: Context,
         ssl: bool,
         client: Optional[Tuple[str, int]],
         server: Optional[Tuple[str, int]],
         send: Callable[[Event], Awaitable[None]],
-        spawn_app: Callable[[dict, Callable], Awaitable[Callable]],
         stream_id: int,
     ) -> None:
+        self.app = app
         self.app_put: Optional[Callable] = None
         self.buffer = WebsocketBuffer(config.websocket_max_message_size)
         self.client = client
         self.closed = False
         self.config = config
+        self.context = context
         self.response: dict
         self.scope: dict
         self.send = send
         # RFC 8441 for HTTP/2 says use http or https, ASGI says ws or wss
         self.scheme = "wss" if ssl else "ws"
         self.server = server
-        self.spawn_app = spawn_app
         self.start_time: float
         self.state = ASGIWebsocketState.HANDSHAKE
         self.stream_id = stream_id
@@ -168,6 +171,7 @@ class WSStream:
 
     async def handle(self, event: Event) -> None:
         if isinstance(event, Request):
+            self.start_time = time()
             self.handshake = Handshake(event.headers, event.http_version)
             path, _, query_string = event.raw_path.partition(b"?")
             self.scope = {
@@ -185,11 +189,15 @@ class WSStream:
                 "subprotocols": self.handshake.subprotocols or [],
                 "extensions": {"websocket.http.response": {}},
             }
-            self.start_time = time()
-            if not self.handshake.is_valid():
+
+            if not valid_server_name(self.config, event):
+                await self._send_error_response(404)
+            elif not self.handshake.is_valid():
                 await self._send_error_response(400)
             else:
-                self.app_put = await self.spawn_app(self.scope, self.app_send)
+                self.app_put = await self.context.spawn_app(
+                    self.app, self.config, self.scope, self.app_send
+                )
                 await self.app_put({"type": "websocket.connect"})
         elif isinstance(event, (Body, Data)):
             self.connection.receive_data(event.data)
@@ -220,16 +228,7 @@ class WSStream:
             await self.send(StreamClosed(stream_id=self.stream_id))
         else:
             if message["type"] == "websocket.accept" and self.state == ASGIWebsocketState.HANDSHAKE:
-                self.state = ASGIWebsocketState.CONNECTED
-                status_code, headers, self.connection = self.handshake.accept(
-                    message.get("subprotocol")
-                )
-                await self.send(
-                    Response(stream_id=self.stream_id, status_code=status_code, headers=headers)
-                )
-                await self.config.log.access(
-                    self.scope, {"status": status_code, "headers": []}, time() - self.start_time
-                )
+                await self._accept(message)
             elif (
                 message["type"] == "websocket.http.response.start"
                 and self.state == ASGIWebsocketState.HANDSHAKE
@@ -301,6 +300,18 @@ class WSStream:
         data = self.connection.send(event)
         await self.send(Data(stream_id=self.stream_id, data=data))
 
+    async def _accept(self, message: dict) -> None:
+        self.state = ASGIWebsocketState.CONNECTED
+        status_code, headers, self.connection = self.handshake.accept(message.get("subprotocol"))
+        await self.send(
+            Response(stream_id=self.stream_id, status_code=status_code, headers=headers)
+        )
+        await self.config.log.access(
+            self.scope, {"status": status_code, "headers": []}, time() - self.start_time
+        )
+        if self.config.websocket_ping_interval is not None:
+            self.context.spawn(self._send_pings)
+
     async def _send_rejection(self, message: dict) -> None:
         body_suppressed = suppress_body("GET", self.response["status"])
         if self.state == ASGIWebsocketState.HANDSHAKE:
@@ -319,3 +330,8 @@ class WSStream:
             self.state = ASGIWebsocketState.HTTPCLOSED
             await self.send(EndBody(stream_id=self.stream_id))
             await self.config.log.access(self.scope, self.response, time() - self.start_time)
+
+    async def _send_pings(self) -> None:
+        while not self.closed:
+            await self._send_wsproto_event(Ping())
+            await self.context.sleep(self.config.websocket_ping_interval)
